@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """Single and multi-threaded executors."""
 import datetime
 import functools
@@ -33,14 +32,15 @@ from .mutation import MutationManager
 from .process import Process, cleanIntermediate, relocateOutputs
 from .provenance_profile import ProvenanceProfile
 from .task_queue import TaskQueue
-from .utils import CWLObjectType, JobsType
+from .update import ORIGINAL_CWLVERSION
+from .utils import CWLObjectType, CWLOutputType, JobsType
 from .workflow import Workflow
 from .workflow_job import WorkflowJob, WorkflowJobStep
 
 TMPDIR_LOCK = Lock()
 
 
-class JobExecutor(object, metaclass=ABCMeta):
+class JobExecutor(metaclass=ABCMeta):
     """Abstract base job executor."""
 
     def __init__(self) -> None:
@@ -89,9 +89,9 @@ class JobExecutor(object, metaclass=ABCMeta):
 
         def check_for_abstract_op(tool: CWLObjectType) -> None:
             if tool["class"] == "Operation":
-                raise SourceLine(tool, "class", WorkflowException).makeError(
-                    "Workflow has unrunnable abstract Operation"
-                )
+                raise SourceLine(
+                    tool, "class", WorkflowException, runtime_context.debug
+                ).makeError("Workflow has unrunnable abstract Operation")
 
         process.visit(check_for_abstract_op)
 
@@ -109,10 +109,7 @@ class JobExecutor(object, metaclass=ABCMeta):
 
         job_reqs = None  # type: Optional[List[CWLObjectType]]
         if "https://w3id.org/cwl/cwl#requirements" in job_order_object:
-            if (
-                process.metadata.get("http://commonwl.org/cwltool#original_cwlVersion")
-                == "v1.0"
-            ):
+            if process.metadata.get(ORIGINAL_CWLVERSION) == "v1.0":
                 raise WorkflowException(
                     "`cwl:requirements` in the input object is not part of CWL "
                     "v1.0. You can adjust to use `cwltool:overrides` instead; or you "
@@ -127,10 +124,7 @@ class JobExecutor(object, metaclass=ABCMeta):
             and "https://w3id.org/cwl/cwl#requirements"
             in cast(CWLObjectType, process.metadata["cwl:defaults"])
         ):
-            if (
-                process.metadata.get("http://commonwl.org/cwltool#original_cwlVersion")
-                == "v1.0"
-            ):
+            if process.metadata.get(ORIGINAL_CWLVERSION) == "v1.0":
                 raise WorkflowException(
                     "`cwl:requirements` in the input object is not part of CWL "
                     "v1.0. You can adjust to use `cwltool:overrides` instead; or you "
@@ -278,7 +272,7 @@ class MultithreadedJobExecutor(JobExecutor):
 
     def __init__(self) -> None:
         """Initialize."""
-        super(MultithreadedJobExecutor, self).__init__()
+        super().__init__()
         self.exceptions = []  # type: List[WorkflowException]
         self.pending_jobs = []  # type: List[JobsType]
         self.pending_jobs_lock = threading.Lock()
@@ -289,35 +283,26 @@ class MultithreadedJobExecutor(JobExecutor):
         self.allocated_cores = float(0)
 
     def select_resources(
-        self, request, runtime_context
-    ):  # pylint: disable=unused-argument
-        # type: (Dict[str, Union[int, float, str]], RuntimeContext) -> Dict[str, Union[int, float, str]]
+        self, request: Dict[str, Union[int, float]], runtime_context: RuntimeContext
+    ) -> Dict[str, Union[int, float]]:  # pylint: disable=unused-argument
         """Naïve check for available cpu cores and memory."""
-        result = {}  # type: Dict[str, Union[int, float, str]]
+        result: Dict[str, Union[int, float]] = {}
         maxrsc = {"cores": self.max_cores, "ram": self.max_ram}
         for rsc in ("cores", "ram"):
             rsc_min = request[rsc + "Min"]
-            if not isinstance(rsc_min, str) and rsc_min > maxrsc[rsc]:
+            if rsc_min > maxrsc[rsc]:
                 raise WorkflowException(
-                    "Requested at least %d %s but only %d available"
-                    % (rsc_min, rsc, maxrsc[rsc])
+                    f"Requested at least {rsc_min} {rsc} but only "
+                    f"{maxrsc[rsc]} available"
                 )
             rsc_max = request[rsc + "Max"]
-            if not isinstance(rsc_max, str) and rsc_max < maxrsc[rsc]:
+            if rsc_max < maxrsc[rsc]:
                 result[rsc] = math.ceil(rsc_max)
             else:
                 result[rsc] = maxrsc[rsc]
 
-        result["tmpdirSize"] = (
-            math.ceil(request["tmpdirMin"])
-            if not isinstance(request["tmpdirMin"], str)
-            else request["tmpdirMin"]
-        )
-        result["outdirSize"] = (
-            math.ceil(request["outdirMin"])
-            if not isinstance(request["outdirMin"], str)
-            else request["outdirMin"]
-        )
+        result["tmpdirSize"] = math.ceil(request["tmpdirMin"])
+        result["outdirSize"] = math.ceil(request["outdirMin"])
 
         return result
 
@@ -332,21 +317,19 @@ class MultithreadedJobExecutor(JobExecutor):
             )
             job.run(runtime_context, TMPDIR_LOCK)
         except WorkflowException as err:
-            _logger.exception("Got workflow error: {}".format(err))
+            _logger.exception(f"Got workflow error: {err}")
             self.exceptions.append(err)
         except Exception as err:  # pylint: disable=broad-except
-            _logger.exception("Got workflow error: {}".format(err))
+            _logger.exception(f"Got workflow error: {err}")
             self.exceptions.append(WorkflowException(str(err)))
         finally:
             if runtime_context.workflow_eval_lock:
                 with runtime_context.workflow_eval_lock:
                     if isinstance(job, JobBase):
                         ram = job.builder.resources["ram"]
-                        if not isinstance(ram, str):
-                            self.allocated_ram -= ram
+                        self.allocated_ram -= ram
                         cores = job.builder.resources["cores"]
-                        if not isinstance(cores, str):
-                            self.allocated_cores -= cores
+                        self.allocated_cores -= cores
                     runtime_context.workflow_eval_lock.notifyAll()
 
     def run_job(
@@ -370,9 +353,7 @@ class MultithreadedJobExecutor(JobExecutor):
                 if isinstance(job, JobBase):
                     ram = job.builder.resources["ram"]
                     cores = job.builder.resources["cores"]
-                    if (not isinstance(ram, str) and ram > self.max_ram) or (
-                        not isinstance(cores, str) and cores > self.max_cores
-                    ):
+                    if ram > self.max_ram or cores > self.max_cores:
                         _logger.error(
                             'Job "%s" cannot be run, requests more resources (%s) '
                             "than available on this host (max ram %d, max cores %d",
@@ -387,11 +368,8 @@ class MultithreadedJobExecutor(JobExecutor):
                         return
 
                     if (
-                        not isinstance(ram, str)
-                        and self.allocated_ram + ram > self.max_ram
-                    ) or (
-                        not isinstance(cores, str)
-                        and self.allocated_cores + cores > self.max_cores
+                        self.allocated_ram + ram > self.max_ram
+                        or self.allocated_cores + cores > self.max_cores
                     ):
                         _logger.debug(
                             'Job "%s" cannot run yet, resources (%s) are not '
@@ -409,11 +387,9 @@ class MultithreadedJobExecutor(JobExecutor):
 
                 if isinstance(job, JobBase):
                     ram = job.builder.resources["ram"]
-                    if not isinstance(ram, str):
-                        self.allocated_ram += ram
+                    self.allocated_ram += ram
                     cores = job.builder.resources["cores"]
-                    if not isinstance(cores, str):
-                        self.allocated_cores += cores
+                    self.allocated_cores += cores
                 self.taskqueue.add(
                     functools.partial(self._runner, job, runtime_context, TMPDIR_LOCK),
                     runtime_context.workflow_eval_lock,
