@@ -8,9 +8,11 @@ import io
 import logging
 import os
 import signal
+import subprocess  # nosec
 import sys
 import time
 import urllib
+import warnings
 from codecs import StreamWriter, getwriter
 from collections.abc import MutableMapping, MutableSequence
 from typing import (
@@ -33,25 +35,20 @@ from typing import (
 import argcomplete
 import coloredlogs
 import pkg_resources  # part of setuptools
-from ruamel import yaml
+import ruamel.yaml
 from ruamel.yaml.comments import CommentedMap, CommentedSeq
+from ruamel.yaml.main import YAML
 from schema_salad.exceptions import ValidationException
-from schema_salad.ref_resolver import (
-    ContextType,
-    FetcherCallableType,
-    Loader,
-    file_uri,
-    uri_file_path,
-)
+from schema_salad.ref_resolver import Loader, file_uri, uri_file_path
 from schema_salad.sourceline import strip_dup_lineno
-from schema_salad.utils import json_dumps
+from schema_salad.utils import ContextType, FetcherCallableType, json_dumps, yaml_no_ts
 
-from . import command_line_tool, workflow
+from . import CWL_CONTENT_TYPES, workflow
 from .argparser import arg_parser, generate_parser, get_default_args
 from .builder import HasReqsHints
 from .context import LoadingContext, RuntimeContext, getdefault
 from .cwlrdf import printdot, printrdf
-from .errors import UnsupportedRequirement, WorkflowException
+from .errors import ArgumentException, UnsupportedRequirement, WorkflowException
 from .executors import JobExecutor, MultithreadedJobExecutor, SingleJobExecutor
 from .load_tool import (
     default_loader,
@@ -63,7 +60,7 @@ from .load_tool import (
     resolve_overrides,
     resolve_tool_uri,
 )
-from .loghandler import _logger, defaultStreamHandler
+from .loghandler import _logger, configure_logging, defaultStreamHandler
 from .mpi import MpiConfig
 from .mutation import MutationManager
 from .pack import pack
@@ -77,7 +74,7 @@ from .process import (
     use_standard_schema,
 )
 from .procgenerator import ProcessGenerator
-from .provenance import ResearchObject
+from .provenance import ResearchObject, WritableBagFile
 from .resolver import ga4gh_tool_registries, tool_resolver
 from .secrets import SecretStore
 from .software_requirements import (
@@ -85,7 +82,7 @@ from .software_requirements import (
     get_container_from_software_requirements,
 )
 from .stdfsaccess import StdFsAccess
-from .subgraph import get_subgraph
+from .subgraph import get_process, get_step, get_subgraph
 from .update import ALLUPDATES, UPDATES
 from .utils import (
     DEFAULT_TMP_PREFIX,
@@ -94,12 +91,10 @@ from .utils import (
     CWLOutputType,
     adjustDirObjs,
     normalizeFilesDirs,
-    onWindows,
     processes_to_kill,
     trim_listing,
     versionstring,
     visit_class,
-    windows_default_container_id,
 )
 from .workflow import Workflow
 
@@ -118,7 +113,22 @@ def _terminate_processes() -> None:
     # It's possible that another thread will spawn a new task while
     # we're executing, so it's not safe to use a for loop here.
     while processes_to_kill:
-        processes_to_kill.popleft().kill()
+        process = processes_to_kill.popleft()
+        cidfile = [
+            str(arg).split("=")[1] for arg in process.args if "--cidfile" in str(arg)
+        ]
+        if cidfile:
+            try:
+                with open(cidfile[0]) as inp_stream:
+                    p = subprocess.Popen(  # nosec
+                        ["docker", "kill", inp_stream.read()], shell=False  # nosec
+                    )
+                    try:
+                        p.wait(timeout=10)
+                    except subprocess.TimeoutExpired:
+                        p.kill()
+            except FileNotFoundError:
+                pass
 
 
 def _signal_handler(signum: int, _: Any) -> None:
@@ -149,10 +159,10 @@ def generate_example_input(
         "float": 0.1,
         "double": 0.1,
         "string": "a_string",
-        "File": yaml.comments.CommentedMap(
+        "File": ruamel.yaml.comments.CommentedMap(
             [("class", "File"), ("path", "a/file/path")]
         ),
-        "Directory": yaml.comments.CommentedMap(
+        "Directory": ruamel.yaml.comments.CommentedMap(
             [("class", "Directory"), ("path", "a/directory/path")]
         ),
     }  # type: CWLObjectType
@@ -165,7 +175,7 @@ def generate_example_input(
             example, comment = generate_example_input(inptype[0], default)
             if optional:
                 if comment:
-                    comment = "{} (optional)".format(comment)
+                    comment = f"{comment} (optional)"
                 else:
                     comment = "optional"
         else:
@@ -184,7 +194,7 @@ def generate_example_input(
                 # array of just an enum then list all the options
                 example = first_item["symbols"]
                 if "name" in first_item:
-                    comment = u'array of type "{}".'.format(first_item["name"])
+                    comment = 'array of type "{}".'.format(first_item["name"])
             else:
                 value, comment = generate_example_input(inptype["items"], None)
                 comment = "array of " + comment
@@ -204,34 +214,34 @@ def generate_example_input(
                 example = symbols[0]
             else:
                 example = "{}_enum_value".format(inptype.get("name", "valid"))
-            comment = u'enum; valid values: "{}"'.format('", "'.join(symbols))
+            comment = 'enum; valid values: "{}"'.format('", "'.join(symbols))
         elif inptype["type"] == "record":
-            example = yaml.comments.CommentedMap()
+            example = ruamel.yaml.comments.CommentedMap()
             if "name" in inptype:
-                comment = u'"{}" record type.'.format(inptype["name"])
+                comment = '"{}" record type.'.format(inptype["name"])
             for field in cast(List[CWLObjectType], inptype["fields"]):
                 value, f_comment = generate_example_input(field["type"], None)
                 example.insert(0, shortname(cast(str, field["name"])), value, f_comment)
         elif "default" in inptype:
             example = inptype["default"]
-            comment = u'default value of type "{}".'.format(inptype["type"])
+            comment = 'default value of type "{}".'.format(inptype["type"])
         else:
             example = defaults.get(cast(str, inptype["type"]), str(inptype))
-            comment = u'type "{}".'.format(inptype["type"])
+            comment = 'type "{}".'.format(inptype["type"])
     else:
         if not default:
             example = defaults.get(str(inptype), str(inptype))
-            comment = u'type "{}"'.format(inptype)
+            comment = f'type "{inptype}"'
         else:
             example = default
-            comment = u'default value of type "{}".'.format(inptype)
+            comment = f'default value of type "{inptype}".'
     return example, comment
 
 
 def realize_input_schema(
-    input_types: MutableSequence[CWLObjectType],
+    input_types: MutableSequence[Union[str, CWLObjectType]],
     schema_defs: MutableMapping[str, CWLObjectType],
-) -> MutableSequence[CWLObjectType]:
+) -> MutableSequence[Union[str, CWLObjectType]]:
     """Replace references to named typed with the actual types."""
     for index, entry in enumerate(input_types):
         if isinstance(entry, str):
@@ -241,32 +251,33 @@ def realize_input_schema(
                 input_type_name = entry
             if input_type_name in schema_defs:
                 entry = input_types[index] = schema_defs[input_type_name]
-        if isinstance(entry, Mapping):
+        if isinstance(entry, MutableMapping):
             if isinstance(entry["type"], str) and "#" in entry["type"]:
                 _, input_type_name = entry["type"].split("#")
                 if input_type_name in schema_defs:
-                    input_types[index]["type"] = cast(
+                    entry["type"] = cast(
                         CWLOutputAtomType,
                         realize_input_schema(
                             cast(
-                                MutableSequence[CWLObjectType],
+                                MutableSequence[Union[str, CWLObjectType]],
                                 schema_defs[input_type_name],
                             ),
                             schema_defs,
                         ),
                     )
             if isinstance(entry["type"], MutableSequence):
-                input_types[index]["type"] = cast(
+                entry["type"] = cast(
                     CWLOutputAtomType,
                     realize_input_schema(
-                        cast(MutableSequence[CWLObjectType], entry["type"]), schema_defs
+                        cast(MutableSequence[Union[str, CWLObjectType]], entry["type"]),
+                        schema_defs,
                     ),
                 )
             if isinstance(entry["type"], Mapping):
-                input_types[index]["type"] = cast(
+                entry["type"] = cast(
                     CWLOutputAtomType,
                     realize_input_schema(
-                        [cast(CWLObjectType, input_types[index]["type"])], schema_defs
+                        [cast(CWLObjectType, entry["type"])], schema_defs
                     ),
                 )
             if entry["type"] == "array":
@@ -275,17 +286,20 @@ def realize_input_schema(
                     if not isinstance(entry["items"], str)
                     else [entry["items"]]
                 )
-                input_types[index]["items"] = cast(
+                entry["items"] = cast(
                     CWLOutputAtomType,
                     realize_input_schema(
-                        cast(MutableSequence[CWLObjectType], items), schema_defs
+                        cast(MutableSequence[Union[str, CWLObjectType]], items),
+                        schema_defs,
                     ),
                 )
             if entry["type"] == "record":
-                input_types[index]["fields"] = cast(
+                entry["fields"] = cast(
                     CWLOutputAtomType,
                     realize_input_schema(
-                        cast(MutableSequence[CWLObjectType], entry["fields"]),
+                        cast(
+                            MutableSequence[Union[str, CWLObjectType]], entry["fields"]
+                        ),
                         schema_defs,
                     ),
                 )
@@ -294,9 +308,12 @@ def realize_input_schema(
 
 def generate_input_template(tool: Process) -> CWLObjectType:
     """Generate an example input object for the given CWL process."""
-    template = yaml.comments.CommentedMap()
-    for inp in realize_input_schema(tool.tool["inputs"], tool.schemaDefs):
-        name = shortname(cast(str, inp["id"]))
+    template = ruamel.yaml.comments.CommentedMap()
+    for inp in cast(
+        List[MutableMapping[str, str]],
+        realize_input_schema(tool.tool["inputs"], tool.schemaDefs),
+    ):
+        name = shortname(inp["id"])
         value, comment = generate_example_input(inp["type"], inp.get("default", None))
         template.insert(0, name, value, comment)
     return template
@@ -319,7 +336,8 @@ def load_job_order(
     if len(args.job_order) == 1 and args.job_order[0][0] != "-":
         job_order_file = args.job_order[0]
     elif len(args.job_order) == 1 and args.job_order[0] == "-":
-        job_order_object = yaml.main.round_trip_load(stdin)
+        yaml = yaml_no_ts()
+        job_order_object = yaml.load(stdin)
         job_order_object, _ = loader.resolve_all(
             job_order_object, file_uri(os.getcwd()) + "/"
         )
@@ -334,7 +352,11 @@ def load_job_order(
             if args.basedir
             else os.path.abspath(os.path.dirname(job_order_file))
         )
-        job_order_object, _ = loader.resolve_ref(job_order_file, checklinks=False)
+        job_order_object, _ = loader.resolve_ref(
+            job_order_file,
+            checklinks=False,
+            content_types=CWL_CONTENT_TYPES,
+        )
 
     if (
         job_order_object is not None
@@ -428,7 +450,7 @@ def init_job_order(
         if _logger.isEnabledFor(logging.DEBUG):
             _logger.debug(
                 "Parsed job order from command line: %s",
-                json_dumps(job_order_object, indent=4),
+                json_dumps(job_order_object, indent=4, default=str),
             )
 
     for inp in process.tool["inputs"]:
@@ -439,10 +461,10 @@ def init_job_order(
                 job_order_object = {}
             job_order_object[shortname(inp["id"])] = inp["default"]
 
-    if job_order_object is None:
+    if len(job_order_object) == 0:
         if process.tool["inputs"]:
             if toolparser is not None:
-                print("\nOptions for {} ".format(args.workflow))
+                print(f"\nOptions for {args.workflow} ")
                 toolparser.print_help()
             _logger.error("")
             _logger.error("Input object required, use --help for details")
@@ -531,7 +553,7 @@ def printdeps(
     elif relative_deps == "cwd":
         base = os.getcwd()
     visit_class(deps, ("File", "Directory"), functools.partial(make_relative, base))
-    stdout.write(json_dumps(deps, indent=4))
+    stdout.write(json_dumps(deps, indent=4, default=str))
 
 
 def prov_deps(
@@ -593,9 +615,9 @@ def print_pack(
     """Return a CWL serialization of the CWL document in JSON."""
     packed = pack(loadingContext, uri)
     if len(cast(Sized, packed["$graph"])) > 1:
-        return json_dumps(packed, indent=4)
+        return json_dumps(packed, indent=4, default=str)
     return json_dumps(
-        cast(MutableSequence[CWLObjectType], packed["$graph"])[0], indent=4
+        cast(MutableSequence[CWLObjectType], packed["$graph"])[0], indent=4, default=str
     )
 
 
@@ -609,31 +631,6 @@ def supported_cwl_versions(enable_dev: bool) -> List[str]:
     return versions
 
 
-def configure_logging(
-    args: argparse.Namespace,
-    stderr_handler: logging.Handler,
-    runtimeContext: RuntimeContext,
-) -> None:
-    rdflib_logger = logging.getLogger("rdflib.term")
-    rdflib_logger.addHandler(stderr_handler)
-    rdflib_logger.setLevel(logging.ERROR)
-    if args.quiet:
-        # Silence STDERR, not an eventual provenance log file
-        stderr_handler.setLevel(logging.WARN)
-    if runtimeContext.debug:
-        # Increase to debug for both stderr and provenance log file
-        _logger.setLevel(logging.DEBUG)
-        stderr_handler.setLevel(logging.DEBUG)
-        rdflib_logger.setLevel(logging.DEBUG)
-    fmtclass = coloredlogs.ColoredFormatter if args.enable_color else logging.Formatter
-    formatter = fmtclass("%(levelname)s %(message)s")
-    if args.timestamps:
-        formatter = fmtclass(
-            "[%(asctime)s] %(levelname)s %(message)s", "%Y-%m-%d %H:%M:%S"
-        )
-    stderr_handler.setFormatter(formatter)
-
-
 def setup_schema(
     args: argparse.Namespace, custom_schema_callback: Optional[Callable[[], None]]
 ) -> None:
@@ -662,13 +659,15 @@ class ProvLogFormatter(logging.Formatter):
 
     def __init__(self) -> None:
         """Use the default formatter with our custom formatstring."""
-        super(ProvLogFormatter, self).__init__("[%(asctime)sZ] %(message)s")
+        super().__init__("[%(asctime)sZ] %(message)s")
 
     def formatTime(
         self, record: logging.LogRecord, datefmt: Optional[str] = None
     ) -> str:
-        formatted_time = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(record.created))
-        with_msecs = "%s,%03f" % (formatted_time, record.msecs)
+        formatted_time = time.strftime(
+            "%Y-%m-%dT%H:%M:%S", time.gmtime(float(record.created))
+        )
+        with_msecs = f"{formatted_time},{record.msecs:03f}"
         return with_msecs
 
 
@@ -676,10 +675,10 @@ def setup_provenance(
     args: argparse.Namespace,
     argsl: List[str],
     runtimeContext: RuntimeContext,
-) -> Optional[int]:
+) -> Union[io.TextIOWrapper, WritableBagFile]:
     if not args.compute_checksum:
         _logger.error("--provenance incompatible with --no-compute-checksum")
-        return 1
+        raise ArgumentException()
     ro = ResearchObject(
         getdefault(runtimeContext.make_fs_access, StdFsAccess)(""),
         temp_prefix_ro=args.tmpdir_prefix,
@@ -688,7 +687,7 @@ def setup_provenance(
     )
     runtimeContext.research_obj = ro
     log_file_io = ro.open_log_file_for_activity(ro.engine_uuid)
-    prov_log_handler = logging.StreamHandler(cast(IO[str], log_file_io))
+    prov_log_handler = logging.StreamHandler(log_file_io)
 
     prov_log_handler.setFormatter(ProvLogFormatter())
     _logger.addHandler(prov_log_handler)
@@ -697,7 +696,7 @@ def setup_provenance(
         # Log cwltool command line options to provenance file
         _logger.info("[cwltool] %s %s", sys.argv[0], " ".join(argsl))
     _logger.debug("[cwltool] Arguments: %s", args)
-    return None
+    return log_file_io
 
 
 def setup_loadingContext(
@@ -705,8 +704,11 @@ def setup_loadingContext(
     runtimeContext: RuntimeContext,
     args: argparse.Namespace,
 ) -> LoadingContext:
+    """Prepare a LoadingContext from the given arguments."""
     if loadingContext is None:
         loadingContext = LoadingContext(vars(args))
+        loadingContext.singularity = runtimeContext.singularity
+        loadingContext.podman = runtimeContext.podman
     else:
         loadingContext = loadingContext.copy()
     loadingContext.loader = default_loader(
@@ -739,13 +741,16 @@ def make_template(
         """Force clean representation of 'null'."""
         return self.represent_scalar("tag:yaml.org,2002:null", "null")
 
-    yaml.representer.RoundTripRepresenter.add_representer(type(None), my_represent_none)
-    yaml.main.round_trip_dump(
+    ruamel.yaml.representer.RoundTripRepresenter.add_representer(
+        type(None), my_represent_none
+    )
+    yaml = YAML()
+    yaml.default_flow_style = False
+    yaml.indent = 4
+    yaml.block_seq_indent = 2
+    yaml.dump(
         generate_input_template(tool),
         sys.stdout,
-        default_flow_style=False,
-        indent=4,
-        block_seq_indent=2,
     )
 
 
@@ -754,7 +759,7 @@ def choose_target(
     tool: Process,
     loadingContext: LoadingContext,
 ) -> Optional[Process]:
-    """Walk the given Workflow and find the process that matches args.target."""
+    """Walk the Workflow, extract the subset matches all the args.targets."""
     if loadingContext.loader is None:
         raise Exception("loadingContext.loader cannot be None")
 
@@ -784,6 +789,75 @@ def choose_target(
     return tool
 
 
+def choose_step(
+    args: argparse.Namespace,
+    tool: Process,
+    loadingContext: LoadingContext,
+) -> Optional[Process]:
+    """Walk the given Workflow and extract just args.single_step."""
+    if loadingContext.loader is None:
+        raise Exception("loadingContext.loader cannot be None")
+
+    if isinstance(tool, Workflow):
+        url = urllib.parse.urlparse(tool.tool["id"])
+        if url.fragment:
+            extracted = get_step(tool, tool.tool["id"] + "/" + args.single_step)
+        else:
+            extracted = get_step(
+                tool,
+                loadingContext.loader.fetcher.urljoin(
+                    tool.tool["id"], "#" + args.single_step
+                ),
+            )
+    else:
+        _logger.error("Can only use --single-step on Workflows")
+        return None
+    if isinstance(loadingContext.loader.idx, MutableMapping):
+        loadingContext.loader.idx[extracted["id"]] = extracted
+        tool = make_tool(extracted["id"], loadingContext)
+    else:
+        raise Exception("Missing loadingContext.loader.idx!")
+
+    return tool
+
+
+def choose_process(
+    args: argparse.Namespace,
+    tool: Process,
+    loadingContext: LoadingContext,
+) -> Optional[Process]:
+    """Walk the given Workflow and extract just args.single_step."""
+    if loadingContext.loader is None:
+        raise Exception("loadingContext.loader cannot be None")
+
+    if isinstance(tool, Workflow):
+        url = urllib.parse.urlparse(tool.tool["id"])
+        if url.fragment:
+            extracted = get_process(
+                tool,
+                tool.tool["id"] + "/" + args.single_process,
+                loadingContext.loader.idx,
+            )
+        else:
+            extracted = get_process(
+                tool,
+                loadingContext.loader.fetcher.urljoin(
+                    tool.tool["id"], "#" + args.single_process
+                ),
+                loadingContext.loader.idx,
+            )
+    else:
+        _logger.error("Can only use --single-process on Workflows")
+        return None
+    if isinstance(loadingContext.loader.idx, MutableMapping):
+        loadingContext.loader.idx[extracted["id"]] = extracted
+        tool = make_tool(extracted["id"], loadingContext)
+    else:
+        raise Exception("Missing loadingContext.loader.idx!")
+
+    return tool
+
+
 def check_working_directories(
     runtimeContext: RuntimeContext,
 ) -> Optional[int]:
@@ -869,13 +943,6 @@ def main(
         else:
             runtimeContext = runtimeContext.copy()
 
-        # If on Windows platform, a default Docker Container is used if not
-        # explicitely provided by user
-        if onWindows() and not runtimeContext.default_container:
-            # This docker image is a minimal alpine image with bash installed
-            # (size 6 mb). source: https://github.com/frol/docker-alpine-bash
-            runtimeContext.default_container = windows_default_container_id
-
         # If caller parsed its own arguments, it may not include every
         # cwltool option, so fill in defaults to avoid crashing when
         # dereferencing them in args.
@@ -883,7 +950,13 @@ def main(
             if not hasattr(args, key):
                 setattr(args, key, val)
 
-        configure_logging(args, stderr_handler, runtimeContext)
+        configure_logging(
+            stderr_handler,
+            args.quiet,
+            runtimeContext.debug,
+            args.enable_color,
+            args.timestamps,
+        )
 
         if args.version:
             print(versionfunc())
@@ -901,8 +974,6 @@ def main(
                 _logger.error("CWL document required, no input file was provided")
                 parser.print_help()
                 return 1
-        if args.relax_path_checks:
-            command_line_tool.ACCEPTLIST_RE = command_line_tool.ACCEPTLIST_EN_RELAXED_RE
 
         if args.ga4gh_tool_registries:
             ga4gh_tool_registries[:] = args.ga4gh_tool_registries
@@ -914,10 +985,13 @@ def main(
 
         setup_schema(args, custom_schema_callback)
 
+        prov_log_stream: Optional[Union[io.TextIOWrapper, WritableBagFile]] = None
         if args.provenance:
             if argsl is None:
                 raise Exception("argsl cannot be None")
-            if setup_provenance(args, argsl, runtimeContext) is not None:
+            try:
+                prov_log_stream = setup_provenance(args, argsl, runtimeContext)
+            except ArgumentException:
                 return 1
 
         loadingContext = setup_loadingContext(loadingContext, runtimeContext, args)
@@ -967,7 +1041,7 @@ def main(
             if loadingContext.loader is None:
                 raise Exception("Impossible code path.")
             processobj, metadata = loadingContext.loader.resolve_ref(uri)
-            processobj = cast(CommentedMap, processobj)
+            processobj = cast(Union[CommentedMap, CommentedSeq], processobj)
             if args.pack:
                 stdout.write(print_pack(loadingContext, uri))
                 return 0
@@ -981,7 +1055,11 @@ def main(
             if args.print_pre:
                 stdout.write(
                     json_dumps(
-                        processobj, indent=4, sort_keys=True, separators=(",", ": ")
+                        processobj,
+                        indent=4,
+                        sort_keys=True,
+                        separators=(",", ": "),
+                        default=str,
                     )
                 )
                 return 0
@@ -992,7 +1070,7 @@ def main(
                 return 0
 
             if args.validate:
-                print("{} is valid CWL.".format(args.workflow))
+                print(f"{args.workflow} is valid CWL.")
                 return 0
 
             if args.print_rdf:
@@ -1023,12 +1101,30 @@ def main(
                 else:
                     tool = ctool
 
+            elif args.single_step:
+                ctool = choose_step(args, tool, loadingContext)
+                if ctool is None:
+                    return 1
+                else:
+                    tool = ctool
+
+            elif args.single_process:
+                ctool = choose_process(args, tool, loadingContext)
+                if ctool is None:
+                    return 1
+                else:
+                    tool = ctool
+
             if args.print_subgraph:
                 if "name" in tool.tool:
                     del tool.tool["name"]
                 stdout.write(
                     json_dumps(
-                        tool.tool, indent=4, sort_keys=True, separators=(",", ": ")
+                        tool.tool,
+                        indent=4,
+                        sort_keys=True,
+                        separators=(",", ": "),
+                        default=str,
                     )
                 )
                 return 0
@@ -1188,10 +1284,7 @@ def main(
                 # Unsetting the Generation from final output object
                 visit_class(out, ("File",), MutationManager().unset_generation)
 
-                if isinstance(out, str):
-                    stdout.write(out)
-                else:
-                    stdout.write(json_dumps(out, indent=4, ensure_ascii=False))
+                stdout.write(json_dumps(out, indent=4, ensure_ascii=False, default=str))
                 stdout.write("\n")
                 if hasattr(stdout, "flush"):
                     stdout.flush()
@@ -1258,7 +1351,10 @@ def main(
                 # Ensure last log lines are written out
                 prov_log_handler.flush()
                 # Underlying WritableBagFile will add the tagfile to the manifest
-                prov_log_handler.stream.close()
+                if prov_log_stream:
+                    prov_log_stream.close()
+                # Why not use prov_log_handler.stream ? That is not part of the
+                # public API for logging.StreamHandler
                 prov_log_handler.close()
             research_obj.close(args.provenance)
 
@@ -1279,9 +1375,25 @@ def find_default_container(
     return default_container
 
 
-def run(*args, **kwargs):
-    # type: (*Any, **Any) -> None
+def windows_check() -> None:
+    """See if we are running on MS Windows and warn about the lack of support."""
+    if os.name == "nt":
+        warnings.warn(
+            "The CWL reference runner (cwltool) no longer supports running "
+            "CWL workflows natively on MS Windows as its previous MS Windows "
+            "support was incomplete and untested. Instead, please see "
+            "https://pypi.org/project/cwltool/#ms-windows-users "
+            "for instructions on running cwltool via "
+            "Windows Subsystem for Linux 2 (WSL2). If don't need to execute "
+            "CWL documents, then you can ignore this warning, but please "
+            "consider migrating to https://pypi.org/project/cwl-utils/ "
+            "for your CWL document processing needs."
+        )
+
+
+def run(*args: Any, **kwargs: Any) -> None:
     """Run cwltool."""
+    windows_check()
     signal.signal(signal.SIGTERM, _signal_handler)
     try:
         sys.exit(main(*args, **kwargs))
