@@ -1,69 +1,37 @@
 """Support for executing Docker containers using the Singularity 2.x engine."""
 
+import logging
 import os
 import os.path
 import re
 import shutil
 import sys
-from distutils import spawn
-from subprocess import (  # nosec
-    DEVNULL,
-    PIPE,
-    Popen,
-    TimeoutExpired,
-    check_call,
-    check_output,
-)
+from subprocess import check_call, check_output  # nosec
 from typing import Callable, Dict, List, MutableMapping, Optional, Tuple, cast
 
 from schema_salad.sourceline import SourceLine
 
 from .builder import Builder
 from .context import RuntimeContext
-from .errors import UnsupportedRequirement, WorkflowException
+from .errors import WorkflowException
 from .job import ContainerCommandLineJob
 from .loghandler import _logger
 from .pathmapper import MapperEnt, PathMapper
-from .utils import (
-    CWLObjectType,
-    create_tmp_dir,
-    docker_windows_path_adjust,
-    ensure_non_writable,
-    ensure_writable,
-)
+from .singularity_utils import singularity_supports_userns
+from .utils import CWLObjectType, create_tmp_dir, ensure_non_writable, ensure_writable
 
-_USERNS = None  # type: Optional[bool]
 _SINGULARITY_VERSION = ""
 
 
-def _singularity_supports_userns() -> bool:
-    global _USERNS  # pylint: disable=global-statement
-    if _USERNS is None:
-        try:
-            hello_image = os.path.join(os.path.dirname(__file__), "hello.simg")
-            result = Popen(  # nosec
-                ["singularity", "exec", "--userns", hello_image, "true"],
-                stderr=PIPE,
-                stdout=DEVNULL,
-                universal_newlines=True,
-            ).communicate(timeout=60)[1]
-            _USERNS = (
-                "No valid /bin/sh" in result
-                or "/bin/sh doesn't exist in container" in result
-            )
-        except TimeoutExpired:
-            _USERNS = False
-    return _USERNS
-
-
 def get_version() -> str:
     global _SINGULARITY_VERSION  # pylint: disable=global-statement
-    if not _SINGULARITY_VERSION:
+    if _SINGULARITY_VERSION == "":
         _SINGULARITY_VERSION = check_output(  # nosec
             ["singularity", "--version"], universal_newlines=True
-        )
+        ).strip()
         if _SINGULARITY_VERSION.startswith("singularity version "):
             _SINGULARITY_VERSION = _SINGULARITY_VERSION[20:]
+        _logger.debug(f"Singularity version: {_SINGULARITY_VERSION}.")
     return _SINGULARITY_VERSION
 
 
@@ -80,6 +48,12 @@ def is_version_3_1_or_newer() -> bool:
     return int(version[0]) >= 4 or (int(version[0]) == 3 and int(version[1]) >= 1)
 
 
+def is_version_3_4_or_newer() -> bool:
+    """Detect if Singularity v3.4+ is available."""
+    version = get_version().split(".")
+    return int(version[0]) >= 4 or (int(version[0]) == 3 and int(version[1]) >= 4)
+
+
 def _normalize_image_id(string: str) -> str:
     return string.replace("/", "_") + ".img"
 
@@ -99,9 +73,7 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
         name: str,
     ) -> None:
         """Builder for invoking the Singularty software container engine."""
-        super(SingularityCommandLineJob, self).__init__(
-            builder, joborder, make_path_mapper, requirements, hints, name
-        )
+        super().__init__(builder, joborder, make_path_mapper, requirements, hints, name)
 
     @staticmethod
     def get_image(
@@ -122,6 +94,8 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
         candidates = []
 
         cache_folder = None
+        debug = _logger.isEnabledFor(logging.DEBUG)
+
         if "CWL_SINGULARITY_CACHE" in os.environ:
             cache_folder = os.environ["CWL_SINGULARITY_CACHE"]
         elif is_version_2_6() and "SINGULARITY_PULLFOLDER" in os.environ:
@@ -218,11 +192,11 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
                     found = True
 
             elif "dockerFile" in dockerRequirement:
-                raise WorkflowException(
-                    SourceLine(dockerRequirement, "dockerFile").makeError(
-                        "dockerFile is not currently supported when using the "
-                        "Singularity runtime for Docker containers."
-                    )
+                raise SourceLine(
+                    dockerRequirement, "dockerFile", WorkflowException, debug
+                ).makeError(
+                    "dockerFile is not currently supported when using the "
+                    "Singularity runtime for Docker containers."
                 )
             elif "dockerLoad" in dockerRequirement:
                 if is_version_3_1_or_newer():
@@ -240,18 +214,19 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
                     check_call(cmd, stdout=sys.stderr)  # nosec
                     found = True
                     dockerRequirement["dockerImageId"] = name
-                raise WorkflowException(
-                    SourceLine(dockerRequirement, "dockerLoad").makeError(
+                else:
+                    raise SourceLine(
+                        dockerRequirement, "dockerLoad", WorkflowException, debug
+                    ).makeError(
                         "dockerLoad is not currently supported when using the "
                         "Singularity runtime (version less than 3.1) for Docker containers."
                     )
-                )
             elif "dockerImport" in dockerRequirement:
-                raise WorkflowException(
-                    SourceLine(dockerRequirement, "dockerImport").makeError(
-                        "dockerImport is not currently supported when using the "
-                        "Singularity runtime for Docker containers."
-                    )
+                raise SourceLine(
+                    dockerRequirement, "dockerImport", WorkflowException, debug
+                ).makeError(
+                    "dockerImport is not currently supported when using the "
+                    "Singularity runtime for Docker containers."
                 )
 
         return found
@@ -268,7 +243,7 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
 
         (e.g. hello-world-latest.{img,sif}).
         """
-        if not bool(spawn.find_executable("singularity")):
+        if not bool(shutil.which("singularity")):
             raise WorkflowException("singularity executable is not available")
 
         if not self.get_image(cast(Dict[str, str], r), pull_image, force_pull):
@@ -283,29 +258,27 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
         runtime: List[str], source: str, target: str, writable: bool = False
     ) -> None:
         runtime.append("--bind")
-        runtime.append(
-            "{}:{}:{}".format(
-                docker_windows_path_adjust(source),
-                docker_windows_path_adjust(target),
-                "rw" if writable else "ro",
-            )
-        )
+        # Mounts are writable by default, so 'rw' is optional and not
+        # supported (due to a bug) in some 3.6 series releases.
+        vol = f"{source}:{target}"
+        if not writable:
+            vol += ":ro"
+        runtime.append(vol)
 
     def add_file_or_directory_volume(
         self, runtime: List[str], volume: MapperEnt, host_outdir_tgt: Optional[str]
     ) -> None:
-        if host_outdir_tgt is not None:
-            # workaround for lack of overlapping mounts in Singularity
-            # revert to daa923d5b0be3819b6ed0e6440e7193e65141052
-            # once https://github.com/sylabs/singularity/issues/1607
-            # is fixed
-            if volume.type == "File":
-                shutil.copy(volume.resolved, host_outdir_tgt)
+        if not volume.resolved.startswith("_:"):
+            if host_outdir_tgt is not None and not is_version_3_4_or_newer():
+                # workaround for lack of overlapping mounts in Singularity <3.4
+                if volume.type == "File":
+                    os.makedirs(os.path.dirname(host_outdir_tgt), exist_ok=True)
+                    shutil.copy(volume.resolved, host_outdir_tgt)
+                else:
+                    shutil.copytree(volume.resolved, host_outdir_tgt)
+                ensure_non_writable(host_outdir_tgt)
             else:
-                shutil.copytree(volume.resolved, host_outdir_tgt)
-            ensure_non_writable(host_outdir_tgt)
-        elif not volume.resolved.startswith("_:"):
-            self.append_volume(runtime, volume.resolved, volume.target)
+                self.append_volume(runtime, volume.resolved, volume.target)
 
     def add_writable_file_volume(
         self,
@@ -314,11 +287,8 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
         host_outdir_tgt: Optional[str],
         tmpdir_prefix: str,
     ) -> None:
-        if host_outdir_tgt is not None:
-            # workaround for lack of overlapping mounts in Singularity
-            # revert to daa923d5b0be3819b6ed0e6440e7193e65141052
-            # once https://github.com/sylabs/singularity/issues/1607
-            # is fixed
+        if host_outdir_tgt is not None and not is_version_3_4_or_newer():
+            # workaround for lack of overlapping mounts in Singularity <3.4
             if self.inplace_update:
                 try:
                     os.link(os.path.realpath(volume.resolved), host_outdir_tgt)
@@ -331,14 +301,21 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
             self.append_volume(runtime, volume.resolved, volume.target, writable=True)
             ensure_writable(volume.resolved)
         else:
-            file_copy = os.path.join(
-                create_tmp_dir(tmpdir_prefix),
-                os.path.basename(volume.resolved),
-            )
-            shutil.copy(volume.resolved, file_copy)
-            # volume.resolved = file_copy
-            self.append_volume(runtime, file_copy, volume.target, writable=True)
-            ensure_writable(file_copy)
+            if host_outdir_tgt:
+                # shortcut, just copy to the output directory
+                # which is already going to be mounted
+                if not os.path.exists(os.path.dirname(host_outdir_tgt)):
+                    os.makedirs(os.path.dirname(host_outdir_tgt))
+                shutil.copy(volume.resolved, host_outdir_tgt)
+                ensure_writable(host_outdir_tgt)
+            else:
+                file_copy = os.path.join(
+                    create_tmp_dir(tmpdir_prefix),
+                    os.path.basename(volume.resolved),
+                )
+                shutil.copy(volume.resolved, file_copy)
+                self.append_volume(runtime, file_copy, volume.target, writable=True)
+                ensure_writable(file_copy)
 
     def add_writable_directory_volume(
         self,
@@ -348,35 +325,46 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
         tmpdir_prefix: str,
     ) -> None:
         if volume.resolved.startswith("_:"):
-            if host_outdir_tgt is not None:
-                new_dir = host_outdir_tgt
-            else:
+            # Synthetic directory that needs creating first
+            if not host_outdir_tgt:
                 new_dir = os.path.join(
                     create_tmp_dir(tmpdir_prefix),
-                    os.path.basename(volume.resolved),
+                    os.path.basename(volume.target),
                 )
-            os.makedirs(new_dir)
+                self.append_volume(runtime, new_dir, volume.target, writable=True)
+                os.makedirs(new_dir)
+                # ^^ Unlike Docker, Singularity won't create directories on demand
+            elif not os.path.exists(host_outdir_tgt):
+                os.makedirs(host_outdir_tgt)
         else:
-            if host_outdir_tgt is not None:
-                # workaround for lack of overlapping mounts in Singularity
-                # revert to daa923d5b0be3819b6ed0e6440e7193e65141052
-                # once https://github.com/sylabs/singularity/issues/1607
-                # is fixed
+            if host_outdir_tgt is not None and not is_version_3_4_or_newer():
+                # workaround for lack of overlapping mounts in Singularity < 3.4
                 shutil.copytree(volume.resolved, host_outdir_tgt)
                 ensure_writable(host_outdir_tgt)
             else:
-                if not self.inplace_update:
-                    dir_copy = os.path.join(
-                        create_tmp_dir(tmpdir_prefix),
-                        os.path.basename(volume.resolved),
+                if self.inplace_update:
+                    self.append_volume(
+                        runtime, volume.resolved, volume.target, writable=True
                     )
-                    shutil.copytree(volume.resolved, dir_copy)
-                    source = dir_copy
-                    # volume.resolved = dir_copy
                 else:
-                    source = volume.resolved
-                self.append_volume(runtime, source, volume.target, writable=True)
-                ensure_writable(source)
+                    if not host_outdir_tgt:
+                        tmpdir = create_tmp_dir(tmpdir_prefix)
+                        new_dir = os.path.join(
+                            tmpdir, os.path.basename(volume.resolved)
+                        )
+                        shutil.copytree(volume.resolved, new_dir)
+                        self.append_volume(
+                            runtime, new_dir, volume.target, writable=True
+                        )
+                    else:
+                        shutil.copytree(volume.resolved, host_outdir_tgt)
+                    ensure_writable(host_outdir_tgt or new_dir)
+
+    def _required_env(self) -> Dict[str, str]:
+        return {
+            "TMPDIR": self.CONTAINER_TMPDIR,
+            "HOME": self.builder.outdir,
+        }
 
     def create_runtime(
         self, env: MutableMapping[str, str], runtime_context: RuntimeContext
@@ -389,33 +377,34 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
             "exec",
             "--contain",
             "--ipc",
+            "--cleanenv",
         ]
-        if _singularity_supports_userns():
+        if singularity_supports_userns():
             runtime.append("--userns")
         else:
             runtime.append("--pid")
+
+        container_HOME: Optional[str] = None
         if is_version_3_1_or_newer():
+            # Remove HOME, as passed in a special way (restore it below)
+            container_HOME = self.environment.pop("HOME")
             runtime.append("--home")
             runtime.append(
                 "{}:{}".format(
-                    docker_windows_path_adjust(os.path.realpath(self.outdir)),
-                    self.builder.outdir,
+                    os.path.realpath(self.outdir),
+                    container_HOME,
                 )
             )
         else:
-            runtime.append("--bind")
-            runtime.append(
-                "{}:{}:rw".format(
-                    docker_windows_path_adjust(os.path.realpath(self.outdir)),
-                    self.builder.outdir,
-                )
-            )
-        runtime.append("--bind")
-        tmpdir = "/tmp"  # nosec
-        runtime.append(
-            "{}:{}:rw".format(
-                docker_windows_path_adjust(os.path.realpath(self.tmpdir)), tmpdir
+            self.append_volume(
+                runtime,
+                os.path.realpath(self.outdir),
+                self.environment["HOME"],
+                writable=True,
             )
+
+        self.append_volume(
+            runtime, os.path.realpath(self.tmpdir), self.CONTAINER_TMPDIR, writable=True
         )
 
         self.add_volumes(
@@ -435,18 +424,18 @@ class SingularityCommandLineJob(ContainerCommandLineJob):
             )
 
         runtime.append("--pwd")
-        runtime.append("%s" % (docker_windows_path_adjust(self.builder.outdir)))
+        runtime.append(self.builder.outdir)
 
-        if runtime_context.custom_net:
-            raise UnsupportedRequirement(
-                "Singularity implementation does not support custom networking"
-            )
-        elif runtime_context.disable_net:
-            runtime.append("--net")
-
-        env["SINGULARITYENV_TMPDIR"] = tmpdir
-        env["SINGULARITYENV_HOME"] = self.builder.outdir
+        if self.networkaccess:
+            if runtime_context.custom_net:
+                runtime.extend(["--net", "--network", runtime_context.custom_net])
+        else:
+            runtime.extend(["--net", "--network", "none"])
 
         for name, value in self.environment.items():
-            env["SINGULARITYENV_{}".format(name)] = str(value)
+            env[f"SINGULARITYENV_{name}"] = str(value)
+
+        if container_HOME:
+            # Restore HOME if we removed it above.
+            self.environment["HOME"] = container_HOME
         return (runtime, None)
