@@ -9,11 +9,9 @@ from cwltool.process import use_custom_schema, use_standard_schema
 from cwltool.update import INTERNAL_VERSION
 from cwltool.utils import CWLObjectType
 
-from .test_fetch import norm
-from .util import get_data, windows_needs_docker
+from .util import get_data
 
 
-@windows_needs_docker
 def test_check_version() -> None:
     """
     It is permitted to load without updating, but not execute.
@@ -23,13 +21,13 @@ def test_check_version() -> None:
     joborder = {"inp": "abc"}  # type: CWLObjectType
     loadingContext = LoadingContext({"do_update": True})
     tool = load_tool(get_data("tests/echo.cwl"), loadingContext)
-    for j in tool.job(joborder, None, RuntimeContext()):
+    for _ in tool.job(joborder, None, RuntimeContext()):
         pass
 
     loadingContext = LoadingContext({"do_update": False})
     tool = load_tool(get_data("tests/echo.cwl"), loadingContext)
     with pytest.raises(WorkflowException):
-        for j in tool.job(joborder, None, RuntimeContext()):
+        for _ in tool.job(joborder, None, RuntimeContext()):
             pass
 
 
@@ -48,17 +46,16 @@ def test_use_metadata() -> None:
 def test_checklink_outputSource() -> None:
     """Is outputSource resolved correctly independent of value of do_validate."""
     outsrc = (
-        norm(Path(get_data("tests/wf/1st-workflow.cwl")).as_uri())
-        + "#argument/classfile"
+        Path(get_data("tests/wf/1st-workflow.cwl")).as_uri() + "#argument/classfile"
     )
 
     loadingContext = LoadingContext({"do_validate": True})
     tool = load_tool(get_data("tests/wf/1st-workflow.cwl"), loadingContext)
-    assert norm(tool.tool["outputs"][0]["outputSource"]) == outsrc
+    assert tool.tool["outputs"][0]["outputSource"] == outsrc
 
     loadingContext = LoadingContext({"do_validate": False})
     tool = load_tool(get_data("tests/wf/1st-workflow.cwl"), loadingContext)
-    assert norm(tool.tool["outputs"][0]["outputSource"]) == outsrc
+    assert tool.tool["outputs"][0]["outputSource"] == outsrc
 
 
 def test_load_graph_fragment() -> None:
@@ -84,7 +81,7 @@ def test_load_graph_fragment_from_packed() -> None:
     loadingContext = LoadingContext()
     uri = Path(get_data("tests/wf/packed-with-loadlisting.cwl")).as_uri() + "#main"
     try:
-        with open(get_data("cwltool/extensions.yml"), "r") as res:
+        with open(get_data("cwltool/extensions.yml")) as res:
             use_custom_schema("v1.0", "http://commonwl.org/cwltool", res.read())
 
         # The updater transforms LoadListingRequirement from an
